%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    Copyright (C) 2004 Adam Spiro and Aharon Bar-Hillel
%    An implementation of a spike sorting algorithm according to the paper:
%    "Spike Sorting: Bayesian Clustering of Non-Stationary Data" 
%    by Aharon Bar-Hillel, Adam Spiro and Eran Stark 
%    (http://www.cs.huji.ac.il/~aharonbh/papers/SpikeSort_NIPS04.ps)
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%  PruneAndSort
%
%   This function prunes the solution and sorts them according to their
%   likelihood. It also merges too small models with the background if it
%   is requested by GParams.SmallClusterThreshold
%
%  Solutions - The solutions to prune and sort
%  LL - The log likelihood.
%  GParams,data - The GParams and the data.
%  N - A vector which contains the sizes of the time frames.
%  prun - a flag == 0 if you only want to sort the data without pruning it.
%  sort - a flag == 1 of you want to sort.
%  partial - the index from which you want to partially prune.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
function [PSolutions, PLL] = PruneAndSort(Solutions,LL,GParams,data,N,prun,sorted,partial);

% Prune and sort script :used after gatherSolutions2

%run over all the windows
pruned=0;       % for prune statistics
stayed=0;
MaxNum=0;       % for sort : keep the largest number of solutions in a row, so additional columns can be removed.

Solutions=reorderBackGround(Solutions);

for j=1:GParams.WinNum
    PrunIndex = []; %keeps the indices of the pruned solutions
    
    if (prun)
        disp(sprintf('Pruning Window %d',j));
        for i=partial:size(Solutions,2)
            if ~isempty(Solutions{j,i})
                KeepFlag=1;
                
                % Add a check that the projection of the x and y axis does
                % not have more than 1 main pick. e.g, take the largest
                % pick and check the amount of spikes in that pick is above
                % 90%. 
                s=0;
                fr=sum(N(1:j-1))+1:sum(N(1:j));
                tmp=confidentAss(Solutions{j,i}, data(fr,:),GParams.EMParams,GParams.ConfidenceMin);
                for label=unique(tmp)
                    if (label>1)
                        inds=find(tmp==label); % the indices of frame j which are equal to label
                        [stmp b]=GaussianSimilarity(data(fr(inds),:));
                        % check if there two high peaks that are seperated
                        % with low valey
                        if (size(b,2)>1)
                            sr=b/max(b)*100;
                            [f g]=sort(sr);
                            m=g(end-1:end);
                            m=sort(m);
                            for v=g([find(f<20)])
                                if (v>m(1) && v<m(2) && size(inds,2)/size(tmp,2) > GParams.ValleyHeight) % if there is a valey
                                    PrunIndex = [PrunIndex i];
                                    pruned=pruned+1;
                                    KeepFlag=0;
                                    break;
                                end
                            end
                        end
                        if (stmp>s)
                            s=stmp;
                            ml=label;
                        end
                    end
                end
                if (KeepFlag==1 && s>GParams.GaussianLikeScore)
 %                  [g h]=uniqFreq(tmp);
 %                   p=h/sum(h);
 %                   ii=find(g==ml);
 %                   if (p(ii)>0.1) % prune only if the pruned cluster is big
                        PrunIndex = [PrunIndex i];
                        pruned=pruned+1;
                        KeepFlag=0;
 %                   end
                end
                
                if (KeepFlag)
                    if (Solutions{j,i}{1,3}>GParams.MaxBGSize) % exclude solutions with large backgrounds
                        PrunIndex = [PrunIndex i];
                        pruned=pruned+1;
                        KeepFlag=0;
                    end
                   
                    for k=1:i-1 % Prun solutions which have similar solutions in the solutions bank
                        if KeepFlag && ~isempty(Solutions{j,k})
                            res= mix_dist3(Solutions{j,i},Solutions{j,k},[N(j) N(j)],GParams); 
                            if -res/(2*N(j)) < GParams.PrunDist
                                KeepFlag=0;
                                PrunIndex = [PrunIndex i];
                                pruned=pruned+1;
                                break;
                            end
                        end
                    end
                end
                
                if KeepFlag         
                    % Look for small clusters and remove them
                     if Solutions{j,i}{1,3}==1 % some bug
                         Solutions{j,i}{1,4}=0;%zeros(1,size(Solutions{j,i},1));
 %                       dbstop in PruneAndSort at 50;
                    end
                    bg=find(cellfun('isempty',Solutions{j,i}(:,4))==1); % the background model index
                    tmp=cell2mat(Solutions{j,i}(:,3));
                    tmp=tmp(setdiff(1:end,bg));
                    exc = find(tmp*100<GParams.SmallClusterThreshold); % clusters to exclude
                    if ~isempty(exc)
                        Solutions{j,i}(bg,3)={cell2mat(Solutions{j,i}(1,3))+sum(tmp(exc))}; % move the weight to the background
                        Solutions{j,i}=Solutions{j,i}(setdiff(1:end,exc+1),:);
                        for k=(setdiff(1:size(Solutions{j,i},1),bg))
                            Solutions{j,i}{k,4}=zeros(1,size(Solutions{j,i},1));
                        end
                        if j==1
                            tData=data(1:N(1),:);
                        else
                            tData=data((N(j-1)+1):sum(N(1:j)),:);
                        end
                        [stam , logLikelihood ]=calc_map_labeling2(tData,Solutions{j,i});
                        LL(j,i) = logLikelihood;
                    end     % of small custers
                    stayed=stayed+1;
                end     % of if keepFlag
            end     % of if ~empty(Solution{j,i})
        end %of i
    end % of if (prun)
    
    
    Ssize = length(Solutions(j,:));
    PSolutions(j,1:Ssize)=Solutions(j,1:Ssize);
    PLL(j,1:Ssize)=LL(j,1:Ssize);
    
    % remove the pruned solutions
    PSolutions(j,PrunIndex) = cell(1,length(PrunIndex));
    PLL(j,PrunIndex) = -Inf;
    
    % permute the solutions so the first ones will be random
    if ~sorted
        perm = randperm(length(PLL(j,:)));
        PLL(j,1:end)=PLL(j,perm);
        PSolutions(j,1:end)=PSolutions(j,perm);
    end
    % just move the zeros to the end and turn them to -Inf
    PLL(j,find((PLL(j,:)==0)))=-Inf;
    pzeros=find(PLL(j,:)==-Inf);
    tmpN=size(PLL,2);
    PLL(j,:)=PLL(j,[setdiff(1:tmpN,pzeros) pzeros]);
    PSolutions(j,1:tmpN)=[PSolutions(j,setdiff(1:tmpN,pzeros)) cell(1,length(pzeros))];
    
    
    if sorted
        % sort the solutions
        [ tPLL , inds ] = sort(PLL(j,:));
        inds=inds(end:-1:1);
        PLL(j,:)=tPLL(end:-1:1);
        PSolutions(j,:)=PSolutions(j,inds);
        localJN=find(isfinite(PLL(j,:)));
        if (~isempty(localJN))
            MaxNum=max(MaxNum,localJN(end));
        else
            MaxNum=0;
        end
    end
end % of j

if prun
    disp(sprintf(' [ pruned stayed]=[ %0.2f %0.2f ]',pruned /(pruned+stayed),  stayed/(pruned+stayed) ));
end

if sorted
    PSolutions=PSolutions(:,1:MaxNum);
    PLL=PLL(:,1:MaxNum);
end