%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    Copyright (C) 2004 Adam Spiro and Aharon Bar-Hillel
%    An implementation of a spike sorting algorithm according to the paper:
%    "Spike Sorting: Bayesian Clustering of Non-Stationary Data" 
%    by Aharon Bar-Hillel, Adam Spiro and Eran Stark 
%    (http://www.cs.huji.ac.il/~aharonbh/papers/SpikeSort_NIPS04.ps)
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% Calculates the scores between every two mixtures in subsequent time
% windows.
% input : Solutions - a cell array of length WinNum, containing mixture solutions 
%               N - a vector (1,WinNum) of number of spikes in each window.
% output : BiMixScores - a cell (1,WinNum) of matrixes. Matrix  i contains logP( h_t-1, h_t), 
%                   where values of h_t-1 are assigned to rows and values of h_t to columns
%                   Correspondences - a cell (1,WinNum) of the correspondences computed 
%                   between every mixture pair. Each cell contains a 3-D array. Entry (i,j,:) 
%                   contains the permutation between mixtures i and j. Entry k in this
%                   (i,j,:) vector contains the index of the correspondent of cluster k from
%                   window t-1 in window t.

function  [BiMixScores ,Correspondences ,Z ] = calc_mix_scores(Solutions,N,GParams)

tmp = size(Solutions);
WinNum = tmp(1);
SolNum = tmp(2);
BiMixScores = cell(WinNum-1,1);
Correspondences=cell(WinNum-1,1);

% Initialize
for k=1:(WinNum-1)
    BiMixScores{k}=ones(SolNum)*(-inf);
end

% calculate the solutions number for each time window
D = cellfun('size',Solutions,1);
if (size(D,2) == 1)
    SolutionsNumber = (D'>0);
else
    SolutionsNumber=sum(D'>0);
end
Z=0;

% calculate the scores
for k=1:(WinNum-1)
    disp(sprintf('Computing transition scores for window %d',k));
    for i=1:SolutionsNumber(k)
        for j=1:SolutionsNumber(k+1)
            % for every two mixtures in windows k and k+1
            % find the distance between these two mixtures
            switch GParams.MethodType
                case 1  % lose at least one bit in labels, encode only Gaussian offsets
                    [ BiMixScores{k}(i,j) tmp ] = mix_dist3(Solutions{k,i},Solutions{k+1,j},...
                        N(k:k+1), GParams);
                case 2  % lose 'exactly' one bit in labels, encode only Gaussian offsets
                    [ BiMixScores{k}(i,j) tmp ] = mix_dist4(Solutions{k,i},Solutions{k+1,j},...
                        N(k:k+1), GParams);
                case 3  % find best labels+offsets model, encode offsets+best model labels
                    [ BiMixScores{k}(i,j) tmp ] = mix_dist2(Solutions{k,i},Solutions{k+1,j},...
                        N(k:k+1), GParams);
                case 4 % like 1, but with a breakdown option
                     [ BiMixScores{k}(i,j) tmp, stam, Z(k,i,j)  ] = mix_dist5(Solutions{k,i},Solutions{k+1,j},...
                        N(k:k+1), GParams);
            end
            % Correspondences{k}(i,j,:)=tmp;
            Correspondences{k,i,j}=tmp;
        end
    end
end