%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%    Copyright (C) 2004 Adam Spiro and Aharon Bar-Hillel
%    An implementation of a spike sorting algorithm according to the paper:
%    "Spike Sorting: Bayesian Clustering of Non-Stationary Data" 
%    by Aharon Bar-Hillel, Adam Spiro and Eran Stark 
%    (http://www.cs.huji.ac.il/~aharonbh/papers/SpikeSort_NIPS04.ps)
%
%    This program is free software; you can redistribute it and/or modify
%    it under the terms of the GNU General Public License as published by
%    the Free Software Foundation; either version 2 of the License, or
%    (at your option) any later version.
%
%    This program is distributed in the hope that it will be useful,
%    but WITHOUT ANY WARRANTY; without even the implied warranty of
%    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
%    GNU General Public License for more details.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


% Calculates the distance between two mixtures, by looking for the most
% probable mixture which includes both. 
% this version tries to solve the constrained maximization
% max P(M1,M2 | R, tetha) ~ exp( -N sum( alpha_i JS(R_i) ) )
% under a constraint H(P_r) <=K*0.5*[ H(alpha_1)+H(alpha_2)]
% the maximization is done in a gready way: at each round we merge two
% clusters which minimize the ratio dJS/dH. the allowed merges are constrained
% by the bipartite.

% inputs :  M1,M2 - two MOG structures to compare
%                   N - a vecotr (1,2) of the sizes of the samples taken from MOG M1,M2 respectively
%                  GParams - uses fields SafConst and BGF - Background First Flag.
%                       BGF - when 1, the backgrounds ( assumed to be 
%                       in places 1 in their mixtures) are united first. When 0 , no special
%                       treatment is given to backgrounds. 
% output : dist - the distance between the two mixtures :  N*sum*alpha_i(JS(R_i))
%                  united - a vector k1+k2 * 1 of the unified labels for the gaussian models. 
%                    Models i,j are together iff united(i)==united(j)
%                  uModel - the unified suggested model.

function [ dist , united , uModel ] = mix_dist3(M1,M2,N,GParams)

SafConst=GParams.SafConst;   % this constant dtermines the bound in the copnstraint over label cost.
% when high, more Gaussian unions are needed and
% the JS returned is lower.

tmp = size(M1);
BlueK=tmp(1);
tmp=size(M2);
RedK=tmp(1);
totN=sum(N);

%BlueK = BlueK-1 % dont' pay for the background model
%RedK = RedK-1;
%totN=(1-M1{end,3})*N(1) + (1-M2{end,3})*N(2);     % total N - number of spikes in the two mixtures.

% prepare a table 2*(number of Gaussians) indexed by the Gaussian number, 
% which says how many blue and red Gaussians  are part of the Gaussian.
CurrM=BlueK+RedK;       % the size (number of Gaussains) of the current mixture.
Color=zeros(CurrM,2);       
Color(1:BlueK,1)=1;
Color(BlueK+1:CurrM,2)=1;

% create the starting pool of Gaussians. 
Gauss=cell(CurrM,5);     % a cell containing the mixture.
Gauss(1:BlueK,1:3)=M1(1:BlueK,1:3);
Gauss(BlueK+1:CurrM,1:3)=M2(1:RedK,1:3);
for i=1:CurrM
    Gauss{i,3}=Gauss{i,3}*(Color(i,:)*N') / totN;
    Gauss{i,4}=det(Gauss{i,2});
    Gauss{i,5}=i;       % the initial gaussians which are part of the Gaussian.
end

% prepare initial distance matrix.
d = ones(CurrM,CurrM)*inf;
for i=1:CurrM
    for j=i+1:CurrM
        Jcolor=Color(i,:)+Color(j,:);
        if any (Jcolor==1)
            d(i,j) =g_dist3(Gauss(i,:),Gauss(j,:));
        end
    end
end

JS=0;       % the JS score between current model and the full double mixture model.
Hloss=0;        % the label entropy gain due to Gaussian unions

% find the first merge
if ~GParams.BGF;         % find the initial best merge
    BestMerge(1)=min(min(d));
    [ a b ]=find(d==BestMerge(1));
    BestMerge(2) = a(1);
    BestMerge(3) = b(1);
else                % start by merging the backgrounds.
    BestMerge(1)=d(1,BlueK+1);
    BestMerge(2) = 1;
    BestMerge(3) = BlueK+1;
end
[ stam , BestMerge(4), BestMerge(5) , newGauss ]= ...
    g_dist3(Gauss(BestMerge(2),:),Gauss(BestMerge(3),:));

while( Hloss<SafConst*log(2) )
    % merge two models and add the merged model
    Gauss(CurrM+1,:)=newGauss;
    Hloss=Hloss+BestMerge(5);
    JS=JS+BestMerge(4);
    
    % compute the new Color and distances  
    Color(CurrM+1,:)=Color(BestMerge(2),:)+Color(BestMerge(3),:);
    d(CurrM+1,1:CurrM+1)=inf;
    for i=1:CurrM       % compute the distances to the new model
        Jcolor=Color(i,:)+Color(CurrM+1,:);
        if any (Jcolor==1)
            d(i,CurrM+1) =g_dist3(Gauss(i,:),Gauss(CurrM+1,:));
        else
            d(i,CurrM+1)= inf;
        end
    end
    
    % remove the old models.
    RemainIndexes=setdiff(1:CurrM+1,BestMerge(2:3));
    Gauss=Gauss( RemainIndexes , : );
    Color=Color(RemainIndexes,:);  
    d=d(RemainIndexes , RemainIndexes);
    CurrM=CurrM-1;
    
    % find the next best merge
    BestMerge(1)=min(min(d));

    if isinf(BestMerge(1))       % no more allowed merges
        break;
    end
    [ I, J ] = find(d==BestMerge(1));
    BestMerge(2)=I(1);
    BestMerge(3)=J(1);
    [ stam , BestMerge(4), BestMerge(5) , newGauss ]= ...
        g_dist3(Gauss(BestMerge(2),:),Gauss(BestMerge(3),:));
    
end

% prepare united
for i=1:CurrM
    united(Gauss{i,5})=i;
end

dist=-totN*JS;
uModel=Gauss(:,1:3);
